/**
 * This data structure maps keys ('IntConst' objects) to values ('IntConst' objects).
 * It is implemented as a binary search tree where keys are ordered according to the order
 * relation of the 'IntConst' class defined by its 'lessThan' and 'isEqual' method.
 * For the keys k1 and k2 of any two mappings in this map the following condition holds:
 * k1.isEqual(k2) == false.
 * There is no limit on the number of key-value mappings stored in the map.
 */
//
// TODO: Complete the methods in 'ConstConstTreeMap'.
//       You can define further classes and methods for the implementation of the
//       binary search tree if needed.
//       Do NOT use the Java-Collection framework in 'ConstConstTreeMap' or any other class.
//
public class ConstConstTreeMap {

    private Node root;

    /**
     * Initializes 'this' as an empty map.
     */
    public ConstConstTreeMap() {
        this.root = null;
    }

    /**
     * Constructs this map as a copy of the specified 'map'. This map has the same key-value mappings
     * as 'map'. If 'map' is changed later, it will not affect 'this' and vice versa.
     *
     * @param map the map from which key-value mappings are copied to this new map, map != null.
     */
    public ConstConstTreeMap(ConstConstTreeMap map) {
        if (map.root == null) return;
        map.root.traversalCopy(this);
    }


    /**
     * Adds a new key-value association to this map. If the key already exists in this map,
     * the value is replaced and the old value is returned. Otherwise, 'null' is returned.
     *
     * @param key   a variable != null.
     * @param value the constant to be associated with the key (can also be 'null').
     * @return the old value if the key already existed in this map, or 'null' otherwise.
     */
    public IntConst put(IntConst key, IntConst value) {
        Node newNode = new Node(key, value);
        if (this.root == null) {
            this.root = newNode;
            return null;
        }

        return this.root.traversalPut(newNode);
    }

    /**
     * Returns the value associated with the specified key. Returns 'null' if the key is not
     * contained in this map.
     *
     * @param key a constant != null.
     * @return the value associated with the specified key (or 'null' if the key is not contained in
     * this map).
     */
    public IntConst get(IntConst key) {
        if (this.root == null) return null;
        return this.root.traversalGet(key);
    }


    /**
     * Calculates the number of keys that are common between this map and another specified map.
     * <p>
     * The operation does not modify either of the maps and ensures that only keys that are present
     * in both maps are considered in the count.
     *
     * @param other The other map to compare with this map, other != null.
     * @return The number of common keys between the two maps.
     */
    public int countCommonKeys(ConstConstTreeMap other) {
        if (this.root == null) return 0;
        return this.root.traversalCount(other);
    }

    /**
     * Returns a string representation of this 'map' with all its mappings in the format: key
     * followed by '=' and the value. The list of mappings is enclosed in square
     * brackets ("[]"). Adjacent elements are separated by ','.
     * The mappings are listed in ascending order according to the key order relation.
     * Example: "[-2=0,1=4,2=4,3=-1,5=5,8=10]"
     *
     * @return the string representation of this map.
     */
    public String toString() {
        if (this.root == null) return "[]";

        StringBuilder builder = new StringBuilder();
        builder.append("[");
        builder.append(this.root.traversalInOrder());
        builder.delete(builder.length() - 1, builder.length());
        builder.append("]");
        return builder.toString();
    }

    /**
     * Returns a string representation of this 'map' listing only those mappings of 'this'
     * with keys lying in the specified inclusive range. The representation has the following
     * format: key followed by '=' and the value. The list of mappings is enclosed in square
     * brackets ("[]"). Adjacent elements are separated by ','.
     * The mappings are listed in ascending order according to the key order relation.
     * Example: "[2=4,3=-1,5=5]"
     * <p>
     * This method exploits the structure of the binary search tree and traverses only
     * relevant parts of the tree.
     *
     * @param lower the inclusive lower bound for the range of keys, lower != null.
     * @param upper the inclusive upper bound for the range of keys, lower <= upper, upper != null.
     * @return the string representation of a range of mappings of this map.
     */
    public String toString(IntConst lower, IntConst upper) {
        if (this.root == null) return "[]";

        StringBuilder builder = new StringBuilder();
        builder.append("[");
        builder.append(this.root.traversalInOrderBoundary(lower, upper));
        builder.delete(builder.length() - 1, builder.length());
        builder.append("]");
        return builder.toString();
    }

}

class Node {
    private Node left;
    private Node right;
    private IntConst key;
    private IntConst value;

    public Node(IntConst key, IntConst value) {
        this.key = key;
        this.value = value;
    }


    public void traversalCopy(ConstConstTreeMap newMap) {
        if (this.left != null) this.left.traversalCopy(newMap);
        newMap.put(this.key, this.value);
        if (this.right != null) this.right.traversalCopy(newMap);

    }

    public IntConst traversalPut(Node nodeToPut) {

        int cmp = this.compare(nodeToPut);

        if (cmp < 0) {
            if (this.right == null) {
                this.right = nodeToPut;
            } else {
                this.right.traversalPut(nodeToPut);
            }
            return null;
        } else if (cmp > 0) {
            if (this.left == null) {
                this.left = nodeToPut;
            } else {
                this.left.traversalPut(nodeToPut);
            }
            return null;
        } else {
            IntConst tmp = this.value;
            this.value = nodeToPut.value;
            return tmp;
        }
    }

    public IntConst traversalGet(IntConst key) {
        int cmp = this.compare(new Node(key, null));
        if (cmp < 0) {
            if (this.right != null) {
                return this.right.traversalGet(key);
            }
        } else if (cmp > 0) {
            if (this.left != null) {
                return this.left.traversalGet(key);
            }
        } else {
            return this.value;
        }
        return null;
    }

    public int traversalCount(ConstConstTreeMap map) {
        Boolean isInMap = map.get(this.key) != null;
        int counter = isInMap ? 1 : 0;
        if (this.left == null && this.right == null) return counter;
        else if (this.left != null && this.right == null) return this.left.traversalCount(map) + counter;
        else if (this.left == null) return this.right.traversalCount(map) + counter;
        else
            return this.right.traversalCount(map) + this.left.traversalCount(map) + counter;

    }

    public String traversalInOrder() {
        String result = "";
        if (this.left != null)
            result += this.left.traversalInOrder();
        result += this.key + "=" + this.value + ",";
        if (this.right != null)
            result += this.right.traversalInOrder();

        return result;

    }

    public String traversalInOrderBoundary(IntConst lower, IntConst upper) {
        String result = "";
        if (this.left != null && (upper.lessThan(this.key) || upper.isEqual(this.key)))
            result += this.left.traversalInOrderBoundary(lower, upper);
        if ((lower.lessThan(this.key) || lower.isEqual(this.key)) && (this.key.isEqual(upper) || this.key.lessThan(upper)))
            result += this.key + "=" + this.value + ",";
        if (this.right != null && (lower.lessThan(this.key) || lower.isEqual(this.key)))
            result += this.right.traversalInOrderBoundary(lower, upper);

        return result;
    }

    private int compare(Node node) {
        if (this.key.lessThan(node.key)) {
            return -1;
        } else if (this.key.isEqual(node.key)) {
            return 0;
        }
        return 1;
    }
}
